/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qdocumentline.h"

/*!
	\file qdocumentline.cpp
	\brief Implementation of the QDocumentLine class.
*/

#include "qdocument_p.h"

/*!
	\ingroup document
	@{
*/

/*!
       \class QDocumentLine

       \brief A reference to line objects

       In QCodeEdit, documents are stored as a list of lines. A QDocumentLine holds
       a pointer to the data of one line and gives access to its content.

       It is not meant to be used to iterate over the document, though it is possible
       for conveneience and compatibility reasons. Indeed, QDocumentLine does not now
	   where in the document it is located. It can obtain that information but this is
	   a O(n) operation. Navigation within the document is one of the task devoted to
       QDocumentCursor which can move around in O(1) (amortized O(1) and O(log(n)) in
	   worst cases).

       Lines can be given formatting in various way : "regular" formatting used for
       highlighting, overlays used mainly to display search matches and similar informations
       and marks.
*/

QDocumentLine::QDocumentLine()
 : m_doc(0), m_handle(0)
{
	
}

QDocumentLine::QDocumentLine(const QDocument *doc)
 : m_doc(doc), m_handle(doc ? doc->impl()->at(0) : 0)
{
	
}

QDocumentLine::QDocumentLine(const QDocumentLine& line)
 : m_doc(line.m_doc), m_handle(line.m_handle)
{
	
}

QDocumentLine::QDocumentLine(const QDocument *doc, QDocumentLineHandle* handle)
 : m_doc(doc), m_handle(handle)
{
	
}

QDocumentLine::~QDocumentLine()
{
	
}

/*!
	\brief Comparision operator
*/
bool QDocumentLine::operator == (const QDocumentLine& l) const
{
	return m_handle == l.m_handle;
	//return lineNumber() == l.lineNumber();
}

/*!
	\brief Comparision operator
*/
bool QDocumentLine::operator != (const QDocumentLine& l) const
{
	return m_handle != l.m_handle;
	//return lineNumber() != l.lineNumber();
}

/*!
	\brief copy operator
	
	QDocumentLine objects are just wrappers around the "real" line data.
	Copies of a QDocumentLine all points to the same underlying data and
	modifying one affect them all (no implicit sharing).
*/
QDocumentLine& QDocumentLine::operator = (const QDocumentLine& l)
{
	m_doc = l.m_doc;
	m_handle = l.m_handle;
	
	return *this;
}

/*!
	\brief Check whether a given flag is set to the line
*/
bool QDocumentLine::hasFlag(State s) const
{
	return m_handle ? m_handle->hasFlag(s) : false;
}

/*!
	\brief Check whether any of the flags in a given combination is set to the line
*/
bool QDocumentLine::hasAnyFlag(int s) const
{
	return m_handle ? m_handle->hasFlag(s) : false;
}

/*!
	\brief Set a flag of the line
	
	\warning Do not mess with flags unless you know what you are doing
*/
void QDocumentLine::setFlag(State s, bool y)
{
	if ( m_handle )
		m_handle->setFlag(s, y);
	
}

/*!
	\return whether the line object is null (i.e invalid)
*/
bool QDocumentLine::isNull() const
{
	return !m_handle;
}

/*!
	\return whether the line object is valid
*/
bool QDocumentLine::isValid() const
{
	return m_handle;
}

/*!
	\return the text of the line
*/
QString QDocumentLine::text() const
{
	return m_handle ? m_handle->text() : QString();
}

/*!
	\return the line number at which this line sit in the document owning it
	
	\note A return value of -1 means that the line is invalid (either "hard"
	invalidity due to null internal pointers or "soft" invalidity if the line
	has been removed from the document by an editing operation but it still
	kept valid for use by the undo/redo stack).
*/
int QDocumentLine::lineNumber() const
{
	return m_doc && m_handle ? m_doc->impl()->indexOf(m_handle) : -1;
}

/*!
	\return The length of the line, in characters
*/
int QDocumentLine::length() const
{
	return m_handle ? m_handle->text().length() : 0;
}

/*!
	\return the number of visual lines occupied by the line
	
	This is NOT set to zero when the line is hidden (e.g due to folding).
*/
int QDocumentLine::lineSpan() const
{
	return (m_doc && m_handle) ? m_doc->impl()->wrapBoundaries(m_handle).count() + 1 : 0;
}

/*!
	\brief Returns the position of the first non-whitespace character
	\return position of first non-whitespace char or -1 if there is none
*/
int QDocumentLine::firstChar() const
{
	return nextNonSpaceChar(0);
}

/*!
	\brief Returns the position of the last non-whitespace character
	\return position of last non-whitespace char or -1 if there is none
*/
int QDocumentLine::lastChar() const
{
	return previousNonSpaceChar(length() - 1);
}

/*!
	\return the indent size in columns (spaces, with tabs instanciated as proper
	number of spaces according to their position and the tabstop)
*/
int QDocumentLine::indent() const
{
	if ( !m_handle || !m_doc )
		return 0;
	
	int l = nextNonSpaceChar(0);
	
	return QDocument::screenLength(text().constData(), l == -1 ? length() : l, m_doc->tabStop());
}

/*!
	Find the position of the next char that is not a space.
	\param pos Column of the character which is examined first.
	\return True if the specified or a following character is not a space
	Otherwise false.
*/
int QDocumentLine::nextNonSpaceChar(int pos) const
{
	return m_handle ? m_handle->nextNonSpaceChar(pos) : -1;
}

/*!
	\brief Find the position of the previous char that is not a space.
	\param pos Column of the character which is examined first.
	\return The position of the first non-whitespace character preceding pos,
   or -1 if none is found.
*/
int QDocumentLine::previousNonSpaceChar(int pos) const
{
	return m_handle ? m_handle->previousNonSpaceChar(pos) : -1;
}

/*!
	\return The wrapped line (i.e "subline") to which a given cursor position resides
	\param cpos cursor position, as a text column
*/
int QDocumentLine::wrappedLineForCursor(int cpos) const
{
	return m_handle && m_doc->impl() ? m_doc->impl()->wrappedLineForCursor(m_handle, cpos) : -1;
}

/*!
	\brief Converts a document offset (viewport) to a cursor position (character / text column)
	
	The (x, y) coordinates given by this function are relative to the absolute
	position of the line, which can be obtained from the document.
*/
int QDocumentLine::documentOffsetToCursor(int x, int y) const
{
	return m_handle && m_doc->impl() ? m_doc->impl()->documentOffsetToCursor(m_handle, x, y) : -1;
}

/*!
	\brief Converts a cursor position (character / text column) to a document offset (viewport)
	
	The (x, y) coordinates given by this function are relative to the absolute
	position of the line, which can be obtained from the document.
*/
void QDocumentLine::cursorToDocumentOffset(int cpos, int& x, int& y) const
{
	if ( m_doc && m_handle )
		m_doc->impl()->cursorToDocumentOffset(m_handle, cpos, x, y);
}

/*!
	\overload
*/
QPoint QDocumentLine::cursorToDocumentOffset(int cpos) const
{
	return m_handle ? m_doc->impl()->cursorToDocumentOffset(m_handle, cpos) : QPoint();
}

/*!
	\brief Toggle a mark on the line
*/
void QDocumentLine::addMark(int id)
{
	if ( !document() )
		return;
	
	document()->impl()->addMark(m_handle, id);
}

/*!
	\brief Toggle a mark on the line
*/
void QDocumentLine::toggleMark(int id)
{
	if ( !document() )
		return;
	
	document()->impl()->toggleMark(m_handle, id);
}

/*!
	\brief Remove a mark from the line
*/
void QDocumentLine::removeMark(int id)
{
	if ( !document() )
		return;
	
	document()->impl()->removeMark(m_handle, id);
}

/*!
	\return the list of marks set on this line
*/
QList<int> QDocumentLine::marks() const
{
	return document() ? document()->impl()->marks(m_handle) : QList<int>();
}

/*!
	\return Whether a given mark has been set on the line
*/
bool QDocumentLine::hasMark(int id) const
{
	return marks().contains(id);
}

/*!
	\return tokens data
	
	\warning You got to check the validity of the line yourself and BEFORE calling this function!
*/
const QVector<int>& QDocumentLine::tokens() const
{
	return m_handle->m_tokens;
}

/*!
	\brief Clear tokens on the line
	
	Meant to be used by language definitions to achieve highlighting. Do not mess with this unless
	you know what you are doing.
*/
void QDocumentLine::clearTokens()
{
	if ( !m_handle )
		return;
	
	m_handle->clearTokens();
}

/*!
	\brief Add a token to the line
	
	Meant to be used by language definitions to achieve highlighting. Do not mess with this unless
	you know what you are doing.
*/
void QDocumentLine::addToken(int index, int length, int tokenId)
{
	if ( !m_handle )
		return;
	
	m_handle->addToken(index, length, tokenId);
}

/*!
	\brief Set the tokens of the line
	
	\note this method is made available for syntax engine use.
	If you want to apply extra formatting on a line use overlays
	instead
	
	\see addOverlay()
*/
void QDocumentLine::setTokens(const QVector<int>& tokens)
{
	if ( !m_handle )
		return;
	
	m_handle->setTokens(tokens);
}

const QVector<int>& QDocumentLine::contexts() const
{
	return m_handle->m_contexts;
}

/*!
	\brief Clear tokens on the line
	
	Meant to be used by language definitions to achieve highlighting. Do not mess with this unless
	you know what you are doing.
*/
void QDocumentLine::clearContexts()
{
	if ( !m_handle )
		return;
	
	m_handle->clearContexts();
}

/*!
	\brief Add a token to the line
	
	Meant to be used by language definitions to achieve highlighting. Do not mess with this unless
	you know what you are doing.
*/
void QDocumentLine::addContext(int index, int length, int contextId)
{
	if ( !m_handle )
		return;
	
	m_handle->addContext(index, length, contextId);
}

/*!
	\brief Set the contexts of the line
	
	\see addContext()
*/
void QDocumentLine::setContexts(const QVector<int>& contexts)
{
	if ( !m_handle )
		return;
	
	m_handle->setContexts(contexts);
}

/*!
	\return whether the line has at least one overlay of a given format id
*/
bool QDocumentLine::hasOverlay(int fid) const
{
	/*
	if ( !m_handle )
		return false;
	
	foreach ( const QFormatRange& r, m_handle->m_overlays )
		if ( r.format == fid )
			return true;
	
	return false;
	*/
	
	return false;
}

/*!
	\brief Clear all overlays applied to the line
*/
QList<QFormatRange> QDocumentLine::overlays() const
{
	/*
	if ( !m_handle )
		return QList<QFormatRange>();
	
	return m_handle->m_overlays;
	*/
	
	return QList<QFormatRange>();
}

/*!
	\brief Clear all overlays applied to the line
*/
void QDocumentLine::clearOverlays()
{
	if ( !m_handle )
		return;
	
	m_handle->clearOverlays();
}

/*!
	\brief Add an overlay to the line
	
	Overlays are format range that get applied on top of regular formatting.
	
	They are typically used to display search matches, matching braces, ...
*/
void QDocumentLine::addOverlay(const QFormatRange& over)
{
	if ( !m_handle )
		return;
	
	m_handle->addOverlay(over);
}

/*!
	\brief Remove an overlay from the line
*/
void QDocumentLine::removeOverlay(const QFormatRange& over)
{
	if ( !m_handle )
		return;
	
	m_handle->removeOverlay(over);
}

/*! @} */
